
/*
 * UNFS3 low-level filehandle routines
 * (C) 2004, Pascal Schmidt <der.eremit@email.de>
 * see file LICENSE for license details
 */

#include <ownfsd.h>

/*
 * hash function for inode numbers
 */
#define FH_HASH(n)  ((n + 3 * (n >> 8) + 5 * (n >> 16)) & 0xFF)

#ifdef OW_MT
    static pthread_mutex_t dircache_mutex = PTHREAD_MUTEX_INITIALIZER ;
    #define DIRCACHELOCK       pthread_mutex_lock(  &dircache_mutex   )
    #define DIRCACHEUNLOCK     pthread_mutex_unlock(&dircache_mutex   )
#else /* OW_MT */
    #define DIRCACHELOCK
    #define DIRCACHEUNLOCK
#endif /* OW_MT */

static int dircache_count = 0 ;
#define DIRCACHE_SIZE 10
struct dircache {
    int verify ;
    struct buspath bp[1] ; /* can extend for longer paths */
} ;

struct dircache * DirCache[DIRCACHE_SIZE] ;

/*
 * --------------------------------
 * FILEHANDLE COMPOSITION FUNCTIONS
 * --------------------------------
 */

/* create a filehandle from a "parsedname" structure */
/* return 0 if ok, else 1 */
int fh_create(char * data_val, const struct parsedname * pn ) {
    unfs3_fh_t * fh = (unfs3_fh_t *) data_val ;
    memcpy( fh->sn,  pn->sn, 8 ) ;
    fh->dev        = pn->dev ;
    fh->ft         = pn->ft ;
    fh->extension  = pn->extension ;
    fh->type       = pn->type ;
    fh->state      = pn->state ;
    fh->pathlength = pn->pathlength ;
    fh->bus_nr     = pn->bus_nr ;
    if ( pn->pathlength>2 ) { /* long path, make an entry into a rotating list */
        int ret = 1 ;
        int i ;
        memcpy( fh->bp, pn->bp, 2*sizeof(struct buspath) ) ;
        DIRCACHELOCK;
            i = dircache_count % DIRCACHE_SIZE ;
            if ( dircache_count > i ) free( DirCache[i] ) ; /* recycling, free first */
            /* allocate space */
            DirCache[i] = malloc( sizeof(struct dircache) + (pn->pathlength - 3 )*sizeof(struct buspath) ) ;
            if (DirCache[i]) {
                memcpy(DirCache[i]->bp, &(pn->bp[2]), (pn->pathlength - 2 )*sizeof(struct buspath) ) ;
                fh->verify = DirCache[i]->verify = dircache_count ;
                fh->pathbuffer = i ;
                ++dircache_count ;
                ret = 0 ;
            }
        DIRCACHEUNLOCK;
        return ret ;
    }
    memcpy( fh->bp, pn->bp, (pn->pathlength)*sizeof(struct buspath) ) ;
    return 0 ;
}

/* return 0 on success */
static int Confirm_device( const struct device * dev, const struct parsedname * pn ) {
    int ret = 1 ;
    void action( const void * t, const VISIT which, const int depth ) {
        (void) depth ;
        if (ret) {
            switch(which) {
            case leaf:
            case postorder:
                if ( dev==(const struct device *)t ) ret = 0 ;
            default:
                break ;
            }
        }
    }
    twalk( Tree[pn->type], action ) ;
    return ret ;
}

static int Confirm_filetype( const struct filetype * ft, const struct device * dev ) {
    int i = dev->nft ;
    for ( ; i>0 ; ) if ( &((dev->ft)[--i]) == ft ) return 0 ;
    return 1 ;
}

static int Add_Branch( char * path, const size_t size, const struct buspath * bp, const struct parsedname * pn ) {
    FS_devicename(path,size,bp->sn,pn) ;
    switch (bp->branch) {
    case 0:
        strcat(path,"/main/") ;
        break ;
    case 1:
        strcat(path,"/aux/") ;
        break ;
    default:
        return 1 ;
    }
    return ( strlen(path) > size ) ;
}

/* recreate a "parsedname" structure from a filehandle */
/* strategy is to check the pointers, then recreate the path string, and finally parse it */
/* return 0 if ok, else 1 */
int fh_interpret(const unfs3_fh_t * fh, struct parsedname * pn ) {
    char path[NFS_MAXPATHLEN+OW_FULLNAME_MAX+1] = "/" ; /* Add space to overwrite the end, then flag an error */
    int start = 1 ;
    /* add state info (uncached, alarm, ... ) */
    /* add path */
    /* confirm type */
    /* confirm device */
    /* add device */
    /* confirm filetype */
    /* add filetype */
    pn->state = fh->state ;
    pn->type = fh->type ;

    start += FS_dirname_state( path, NFS_MAXPATHLEN-start, pn ) ;
    if ( path[start-1] != '/' ) {
        path[start] = '/' ;
        ++start ;
        if ( start >= NFS_MAXPATHLEN ) return 1 ;
    }

    start += FS_dirname_type( &path[start],  NFS_MAXPATHLEN-start, pn ) ;
    if ( path[start-1] != '/' ) {
        path[start] = '/' ;
        ++start ;
        if ( start >= NFS_MAXPATHLEN ) return 1 ;
    }

    if ( fh->pathlength > 0 ) {
        if ( Add_Branch( &path[start], NFS_MAXPATHLEN-start, &(fh->bp[0]), pn ) ) return 1 ;
        start = strlen(path) ;
    }
    if ( fh->pathlength > 1 ) {
        if ( Add_Branch( &path[start], NFS_MAXPATHLEN-start, &(fh->bp[1]), pn ) ) return 1 ;
        start = strlen(path) ;
    }
    if ( fh->pathlength > 2 ) {
        int ret = 0 ;
        int pb = fh->pathbuffer ;
        DIRCACHELOCK;
            if ( (pb >= DIRCACHE_SIZE) || (pb >= dircache_count) || (DirCache[pb]->verify != fh->verify) ) {
                ret = 1 ;
            } else {
                int i ;
                for ( i=2 ; i<fh->pathlength ; ++i ) {
                    if ( Add_Branch( &path[start], NFS_MAXPATHLEN-start, &(DirCache[pb]->bp[i-2]), pn ) ) {
                        ret = 1 ;
                        break ;
                    }
                    start = strlen(path) ;
                }
            }
        DIRCACHEUNLOCK;
        if ( ret ) return 1 ;
    }
    if ( fh->dev ) { /* Any device? */
        switch (fh->type) {
        case pn_real:
            if ( fh->dev == DeviceSimultaneous) {
                strcpy(&path[start],"simultaneous") ;
            } else if ( fh->dev==&NoDevice || Confirm_device( fh->dev, pn )==0 ) {
                FS_devicename( &path[start],NFS_MAXPATHLEN-start, fh->sn, pn ) ;
            } else {
                return 1 ;
            }
            break ;
        case pn_statistics:
        case pn_system:
        case pn_settings:
        case pn_structure:
            if ( Confirm_device(fh->dev, pn) ) return 1 ;
            strcpy( &path[start], fh->dev->code ) ;
            break ;
        default:
            return 1 ;
        }
        start = strlen(path) ;
        if ( start >= NFS_MAXPATHLEN ) return 1 ;
        if ( fh->ft ) {
            if ( Confirm_filetype( fh->ft, fh->dev ) ) return 1 ;
            path[start]='/' ;
            ++start ;
            pn->ft = fh->ft ; /* Load for FS_FileName */
            if ( FS_FileName( &path[start], NFS_MAXPATHLEN-start, pn ) ) return 1 ;
            if ( strlen(path) > NFS_MAXPATHLEN ) return 1 ;
        }
    }
    return FS_ParsedName( path, pn ) ;
}
