
/*
 * UNFS3 NFS protocol procedures
 * (C) 2004, Pascal Schmidt <der.eremit@email.de>
 * see file LICENSE for license details
 */

// #include "config.h"
#include <ownfsd.h>
#include <argz.h>

void nfsproc3_getattr_3_svc(GETATTR3args * argp, struct datapasser_t * dp ) {
    GETATTR3res * result = (GETATTR3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->object.data.data_val ;
    
    if ( fh_interpret( fh, dp->pn ) ) {
        result->status = NFS3ERR_STALE ;
    } else {
        post_op_attr attr ;
        get_post_attr( &attr, dp->pn ) ;
        memcpy( &(result->GETATTR3res_u.resok.obj_attributes), &(attr.post_op_attr_u.attributes), sizeof(fattr3) ) ;
        result->status = NFS3_OK ;
    }
}

void nfsproc3_setattr_3_svc(SETATTR3args * argp, struct datapasser_t * dp ) {
    SETATTR3res * result = (SETATTR3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->object.data.data_val ;
    
    if ( fh_interpret( fh, dp->pn ) ) {
        bad_pre_attr( &(result->SETATTR3res_u.resok.obj_wcc.before) ) ;
        bad_post_attr( &(result->SETATTR3res_u.resok.obj_wcc.after ) ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        get_pre_attr( &(result->SETATTR3res_u.resok.obj_wcc.before), dp->pn );
        get_post_attr( &(result->SETATTR3res_u.resok.obj_wcc.after ), dp->pn );
        result->status = NFS3ERR_NOTSUPP ;
    }
}

void nfsproc3_lookup_3_svc(LOOKUP3args * argp, struct datapasser_t * dp ) {
    LOOKUP3res * result = (LOOKUP3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->what.dir.data.data_val ;
    
    bad_post_attr( &(result->LOOKUP3res_u.resok.dir_attributes) ) ;
    if ( fh_interpret( fh, dp->pn ) ) {
        result->status = NFS3ERR_STALE ;
    } else {
        char *path ;
        size_t len = strlen( dp->pn->path ) + strlen(argp->what.name) + 1 ;
        /* covers ok and fail */
        get_post_attr( &(result->LOOKUP3res_u.resfail.dir_attributes) , dp->pn ) ;
        if ( len >= NFS_MAXPATHLEN ) {
            result->status = NFS3ERR_NAMETOOLONG;
        } else if ( (path=malloc(len+1)) == NULL ) {
            result->status = NFS3ERR_SERVERFAULT ;
        } else if ( (dp->buffer=malloc(sizeof(unfs3_fh_t))) == NULL ) {
            result->status = NFS3ERR_SERVERFAULT ;
        } else {
            struct parsedname pn2 ;
            pn2.si = dp->pn->si ;
            strcpy(path,dp->pn->path) ;
            strcat(path,"/") ;
            strcat(path,argp->what.name) ;
            if ( FS_ParsedName(path,&pn2) || fh_create(dp->buffer, &pn2) ) {
                result->status = NFS3ERR_NOENT;
            } else {
                result->status = NFS3_OK ;
                result->LOOKUP3res_u.resok.object.data.data_len = sizeof(struct nfs_fh3) ;
                result->LOOKUP3res_u.resok.object.data.data_val = dp->buffer;
                get_post_attr( &(result->LOOKUP3res_u.resok.obj_attributes), &pn2 );
            }
            FS_ParsedName_destroy( &pn2 ) ;
        }
    }
}

void nfsproc3_access_3_svc(ACCESS3args * argp, struct datapasser_t * dp ) {
    ACCESS3res * result = (ACCESS3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->object.data.data_val ;
    
    if ( fh_interpret( fh, dp->pn ) ) {
        result->status = NFS3ERR_STALE ;
    } else {
        mode_t mode ;
        uint32 nfsaccess = 0 ;
        
        get_post_attr( &(result->ACCESS3res_u.resok.obj_attributes), dp->pn ) ;
        mode = result->ACCESS3res_u.resok.obj_attributes.post_op_attr_u.attributes.mode ;
        
        if (mode & S_IRUSR) nfsaccess |= ACCESS3_READ;
        if (mode & S_IWUSR) nfsaccess |= ACCESS3_MODIFY | ACCESS3_EXTEND;
        /* adjust if directory */
        if (result->ACCESS3res_u.resok.obj_attributes.post_op_attr_u.attributes.type == NF3DIR)
            nfsaccess |= ACCESS3_LOOKUP & ACCESS3_MODIFY ;

        result->status = NFS3_OK ;
        result->ACCESS3res_u.resok.access = nfsaccess & argp->access;
    }
}

void nfsproc3_readlink_3_svc(READLINK3args * argp, struct datapasser_t * dp ) {
    READLINK3res * result = (READLINK3res *) dp->result ;
    (void) argp ;
    
    result->status = NFS3ERR_NOTSUPP ;
    bad_post_attr( &(result->READLINK3res_u.resfail.symlink_attributes) ) ;
}

/* Read routine --
   resizes results space for read buffer at end
   tries full length, then 4096 if no possible
   returns pointer to reallocated space for freeing
*/    
void nfsproc3_read_3_svc(READ3args * argp, struct datapasser_t * dp ) {
    READ3res * result = (READ3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->file.data.data_val ;
        
    if ( fh_interpret( fh, dp->pn ) ) {
        result->status = NFS3ERR_BADHANDLE ;
    } else {
        /* overlaps with resfail */
        get_post_attr(&(result->READ3res_u.resok.file_attributes) , dp->pn ) ;
        /* is this a valid regular file? */
        if ( result->READ3res_u.resok.file_attributes.attributes_follow &&
             result->READ3res_u.resok.file_attributes.post_op_attr_u.attributes.type == NF3REG ) {
            size_t size = argp->count ;
            off_t offset = argp->offset ;
            size_t count = 4096 ;
            dp->buffer = malloc( size ) ;
            /* Two tries at allocating a read buffer */
            if ( dp->buffer==NULL && size > count ) {
                dp->buffer = malloc( count ) ;
            } else { /* success with -size- */
                count = size ;
            }
            if ( dp->buffer == NULL ) { /* Could not allocate */
                result->status = NFS3ERR_SERVERFAULT ; /* cannot allocate memory */
            } else {
                /* Allocation succeeded, try read */
                int ret = FS_read_3times(dp->buffer,count,offset,dp->pn) ;
                if ( ret >= 0 ) {
                    result->status = NFS3_OK ;
                    result->READ3res_u.resok.count = ret ;
                    result->READ3res_u.resok.eof = (ret+offset==result->READ3res_u.resok.file_attributes.post_op_attr_u.attributes.size ) ;
                    result->READ3res_u.resok.data.data_len = count ;
                    result->READ3res_u.resok.data.data_val = dp->buffer ;
                } else if (ret==-ENOENT) { /* bad read */
                    result->status = NFS3ERR_NXIO ;
                } else { /* bad read */
                    result->status = NFS3ERR_INVAL ;
                }
            }
        } else { /* wrong file type or file andle failure */
            result->status = NFS3ERR_INVAL ;
        }
    }
}

void nfsproc3_write_3_svc(WRITE3args * argp, struct datapasser_t * dp ) {
    WRITE3res * result = ( WRITE3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->file.data.data_val ;
    
    if ( fh_interpret( fh, dp->pn ) ) {
        result->status = NFS3ERR_STALE ;
    } else {
        get_pre_attr( &(result->WRITE3res_u.resfail.file_wcc.before), dp->pn ) ;
        bad_post_attr( &(result->WRITE3res_u.resfail.file_wcc.after ) ) ;
        if ( dp->pn->dev==NULL || dp->pn->ft == NULL ) {
            result->status = NFS3ERR_ISDIR ;
        } else if (dp->pn->type == pn_structure ) { /* structure is read-only */
            result->status = NFS3ERR_IO ;
        } else if ( dp->pn->in == NULL ) {
            result->status = NFS3ERR_IO ;
        } else {
            size_t count = argp->data.data_len ;
            int ret ;

            if ( argp->count < count ) count = argp->count ; /* smaller of data buffer and intended write size */
            ret = FS_write_postparse( argp->data.data_val, count, argp->offset, dp->pn ) ;
            if ( ret >= 0 ) { /* good read */
                result->status = NFS3_OK ;
                result->WRITE3res_u.resok.count = ret ;
                result->WRITE3res_u.resok.committed = FILE_SYNC ;
                memset( &(result->WRITE3res_u.resok.verf), 0, NFS3_WRITEVERFSIZE ) ;
                get_post_attr( &(result->WRITE3res_u.resfail.file_wcc.after ), dp->pn ) ;
            } else if (ret==-EROFS) {
                result->status = NFS3ERR_ROFS ;
            } else {
                result->status = NFS3ERR_INVAL ;
            }
        }
    }
}                

void nfsproc3_create_3_svc(CREATE3args * argp, struct datapasser_t * dp ) {
    CREATE3res * result = (CREATE3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->where.dir.data.data_val ;
    
    if ( fh_interpret( fh, dp->pn ) ) {
        bad_pre_attr( &(result->CREATE3res_u.resfail.dir_wcc.before) ) ;
        bad_post_attr( &(result->CREATE3res_u.resfail.dir_wcc.after ) ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        get_pre_attr( &(result->CREATE3res_u.resfail.dir_wcc.before), dp->pn );
        get_post_attr( &(result->CREATE3res_u.resfail.dir_wcc.after ), dp->pn );
        result->status = NFS3ERR_NOTSUPP ;
    }
}

void nfsproc3_mkdir_3_svc(MKDIR3args * argp, struct datapasser_t * dp ) {
    MKDIR3res * result = (MKDIR3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->where.dir.data.data_val ;
    
    if ( fh_interpret( fh, dp->pn ) ) {
        bad_pre_attr( &(result->MKDIR3res_u.resfail.dir_wcc.before) ) ;
        bad_post_attr( &(result->MKDIR3res_u.resfail.dir_wcc.after ) ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        get_pre_attr( &(result->MKDIR3res_u.resfail.dir_wcc.before), dp->pn );
        get_post_attr( &(result->MKDIR3res_u.resfail.dir_wcc.after ), dp->pn );
        result->status = NFS3ERR_NOTSUPP ;
    }
}

void nfsproc3_symlink_3_svc(SYMLINK3args * argp, struct datapasser_t * dp ) {
    SYMLINK3res * result = (SYMLINK3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->where.dir.data.data_val ;
    
    if ( fh_interpret( fh, dp->pn ) ) {
        bad_pre_attr( &(result->SYMLINK3res_u.resfail.dir_wcc.before) ) ;
        bad_post_attr( &(result->SYMLINK3res_u.resfail.dir_wcc.after ) ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        get_pre_attr( &(result->SYMLINK3res_u.resfail.dir_wcc.before), dp->pn );
        get_post_attr( &(result->SYMLINK3res_u.resfail.dir_wcc.after ), dp->pn );
        result->status = NFS3ERR_NOTSUPP ;
    }
}

void nfsproc3_mknod_3_svc(MKNOD3args * argp, struct datapasser_t * dp ) {
    MKNOD3res * result = (MKNOD3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->where.dir.data.data_val ;
    
    if ( fh_interpret( fh, dp->pn ) ) {
        bad_pre_attr( &(result->MKNOD3res_u.resfail.dir_wcc.before) ) ;
        bad_post_attr( &(result->MKNOD3res_u.resfail.dir_wcc.after ) ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        get_pre_attr( &(result->MKNOD3res_u.resfail.dir_wcc.before), dp->pn );
        get_post_attr( &(result->MKNOD3res_u.resfail.dir_wcc.after ), dp->pn );
        result->status = NFS3ERR_NOTSUPP ;
    }
}

void nfsproc3_remove_3_svc(REMOVE3args * argp, struct datapasser_t * dp ) {
    REMOVE3res * result = (REMOVE3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->object.dir.data.data_val ;
    
    if ( fh_interpret( fh, dp->pn ) ) {
        bad_pre_attr( &(result->REMOVE3res_u.resfail.dir_wcc.before) ) ;
        bad_post_attr( &(result->REMOVE3res_u.resfail.dir_wcc.after ) ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        get_pre_attr( &(result->REMOVE3res_u.resfail.dir_wcc.before), dp->pn );
        get_post_attr( &(result->REMOVE3res_u.resfail.dir_wcc.after ), dp->pn );
        result->status = NFS3ERR_NOTSUPP ;
    }
}

void nfsproc3_rmdir_3_svc(RMDIR3args * argp, struct datapasser_t * dp ) {
    RMDIR3res * result = (RMDIR3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->object.dir.data.data_val ;
    
    if ( fh_interpret( fh, dp->pn ) ) {
        bad_pre_attr( &(result->RMDIR3res_u.resfail.dir_wcc.before) ) ;
        bad_post_attr( &(result->RMDIR3res_u.resfail.dir_wcc.after ) ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        get_pre_attr( &(result->RMDIR3res_u.resfail.dir_wcc.before), dp->pn );
        get_post_attr( &(result->RMDIR3res_u.resfail.dir_wcc.after ), dp->pn );
        result->status = NFS3ERR_NOTSUPP ;
    }
}

void nfsproc3_rename_3_svc(RENAME3args * argp, struct datapasser_t * dp ) {
    RENAME3res * result = (RENAME3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->from.dir.data.data_val ;

    bad_pre_attr( &(result->RENAME3res_u.resfail.todir_wcc.before) ) ;
    bad_post_attr( &(result->RENAME3res_u.resfail.todir_wcc.after ) ) ;
    if ( fh_interpret( fh, dp->pn ) ) {
        bad_pre_attr( &(result->RENAME3res_u.resfail.fromdir_wcc.before) ) ;
        bad_post_attr( &(result->RENAME3res_u.resfail.fromdir_wcc.after ) ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        get_pre_attr( &(result->RENAME3res_u.resfail.fromdir_wcc.before), dp->pn );
        get_post_attr( &(result->RENAME3res_u.resfail.fromdir_wcc.after ), dp->pn );
        result->status = NFS3ERR_NOTSUPP ;
    }
}

void nfsproc3_link_3_svc(LINK3args * argp, struct datapasser_t * dp ) {
    LINK3res * result = (LINK3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->file.data.data_val ;
    
    bad_pre_attr( &(result->LINK3res_u.resfail.linkdir_wcc.before) ) ;
    bad_post_attr( &(result->LINK3res_u.resfail.linkdir_wcc.after ) ) ;
    if ( fh_interpret( fh, dp->pn ) ) {
        bad_post_attr( &(result->LINK3res_u.resfail.file_attributes) ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        get_post_attr( &(result->LINK3res_u.resfail.file_attributes), dp->pn );
        result->status = NFS3ERR_NOTSUPP ;
    }
}    


/*
 * static READDIR3resok size with XDR overhead
 *
 * 88 bytes attributes, 8 bytes verifier, 4 bytes value_follows for
 * first entry, 4 bytes eof flag
 */
#define RESOK_SIZE 104

/*
 * static entry3 size with XDR overhead
 *
 * 8 bytes fileid, 4 bytes name length, 8 bytes cookie, 4 byte value_follows
 */
#define ENTRY_SIZE 24
#define ENTRYPLUS_SIZE (24+64+30)

/*
 * size of a name with XDR overhead
 *
 * XDR pads to multiple of 4 bytes
 */
#define NAME_SIZE(x) (((strlen((x))+3)/4)*4)

#ifdef OW_MT
    static pthread_mutex_t dirbuffer_mutex = PTHREAD_MUTEX_INITIALIZER ;
    static pthread_mutex_t dirscan_mutex   = PTHREAD_MUTEX_INITIALIZER ;
#define DIRBUFFERLOCK       pthread_mutex_lock(  &dirbuffer_mutex   )
#define DIRBUFFERUNLOCK     pthread_mutex_unlock(&dirbuffer_mutex   )
#define DIRSCANLOCK       pthread_mutex_lock(  &dirscan_mutex   )
#define DIRSCANUNLOCK     pthread_mutex_unlock(&dirscan_mutex   )
#else /* OW_MT */
#define DIRBUFFERLOCK
#define DIRBUFFERUNLOCK
#define DIRSCANLOCK
#define DIRSCANUNLOCK
#endif /* OW_MT */

static struct volatilebuffer {
    char *  argz ;
    size_t  argz_len;
    cookie3 cookie;
    char *  lastused ;
} vb = { NULL, 0, 0, NULL, } ;

void nfsproc3_readdir_3_svc(READDIR3args * argp, struct datapasser_t * dp ) {
    READDIR3res * result = (READDIR3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->dir.data.data_val ;
    size_t sbuf = argp->count ;
    int indx = 0 ;
    int eof = 1 ;
    int spacetogo;
    char * endadds ;
    int nentry ;

    /* Embedded function */
    /* Add an entry to the allocated buffer */
    /* structs at front, strings at end */
    /* Needs to be called under mutex protection */
    int AddDirEntry( char * name, cookie3 ck ) {
        entry3 * E = (entry3 *) dp->buffer ;
        int l = strlen(name) ;
        int ret = 0 ;

        DIRSCANLOCK;
            /* test C length */
            if ( dp->buffer + (indx+1)*sizeof(entry3) + l + 1 > endadds ) {
                ret = 1 ; /* buffer full */
            } else {
                /* test XDR length */
                spacetogo -= ENTRY_SIZE + ((l+3)/4)*4 ;
                if ( spacetogo < 0 ) {
                    ret = 1 ;
                } else {
                    endadds -= l+1 ;
                    strcpy(endadds,name) ;
            
                    if ( indx>0 ) E[indx-1].nextentry = &(E[indx]) ;
                    E[indx].cookie = ck ;
                    E[indx].fileid = UniqueNum() ;
                    E[indx].name = endadds ;
                    E[indx].nextentry = 0 ;
                    ++ indx ;
                }
            }
        DIRSCANUNLOCK;
        return ret ;
    }

    /* Embedded function */
    /* Callback function to FS_dir */
    /* Prints this directory element (not the whole path) */
    void load_volatile( const struct parsedname * const pn2 ) {
        char extname[OW_FULLNAME_MAX+1] ;
        FS_DirName( extname, OW_FULLNAME_MAX+1, pn2 ) ;
        if ( AddDirEntry( extname, (cookie3) 1 ) ) {
            /* Buffer full, now store rest of entries */
            DIRBUFFERLOCK;
                if ( eof ) { /* first excess entry */
                    cookie3 cook = ++vb.cookie ;
                    memcpy( &(result->READDIR3res_u.resok.cookieverf), &cook, sizeof(cookie3) ) ;
                    if ( vb.argz ) {
                        free(vb.argz) ;
                        vb.argz = NULL ;
                        vb.argz_len = 0 ;
                        vb.lastused = 0 ;
                        eof = 0 ;
                    }
                }
                argz_add( &(vb.argz), &(vb.argz_len), extname ) ;
            DIRBUFFERLOCK;
        }
    }

    /* Embedded function */
    void next_volatile( void ) {
        char * next ;
        while ( (next=argz_next( vb.argz, vb.argz_len, vb.lastused)) ) {
            if ( AddDirEntry( next, (cookie3) 2 ) ) return ;
            vb.lastused = next ;
        }
        eof = 1 ;
    }           

    /* Embedded function */
    /* Callback function to FS_dir */
    /* Prints this directory element (not the whole path) */
    void static_dir( const struct parsedname * const pn2 ) {
       char extname[OW_FULLNAME_MAX+1] ;
        if ( (++nentry > argp->cookie) && eof) {
            FS_DirName( extname, OW_FULLNAME_MAX+1, pn2 ) ;
            if ( AddDirEntry( extname, (cookie3) nentry ) ) eof = 0 ;
        }
    }
    
    bad_post_attr( &(result->READDIR3res_u.resfail.dir_attributes) ) ;
    if ( sbuf < 64 ) {
        result->status = NFS3ERR_TOOSMALL ;
        return ;
    }
    
    if ( sbuf>4096) sbuf=4096 ; /* limit buffer size */

    if ( (dp->buffer = malloc(sbuf)) == NULL ) {
        result->status = NFS3ERR_SERVERFAULT ;
        return ;
    }
    spacetogo = argp->count - sizeof(READDIR3resok) ;
    endadds = &(dp->buffer[sbuf]) ;

    if ( fh_interpret( fh, dp->pn ) ) {
       result->status = NFS3ERR_STALE ;
    } else {
        get_post_attr( &(result->READDIR3res_u.resfail.dir_attributes), dp->pn );
        switch( dp->pn->desc ) {
        case desc_error:
            result->status = NFS3ERR_IO ;
            break ;
        case desc_alarmdir:
        case desc_rootdir:
        case desc_branchdir:
            /* changable directory -- load buffer, then store excess for next call */
            switch( argp->cookie ) {
            case 1:
                /* not first pass */
                DIRBUFFERLOCK;
                    if ( memcmp( argp->cookieverf, &(vb.cookie), sizeof(cookie3) ) ) {
                        result->status = NFS3ERR_BAD_COOKIE ;
                    } else {
                        eof = 0 ;
                        memcpy( &(result->READDIR3res_u.resok.cookieverf), &(vb.cookie), sizeof(cookie3) ) ;
                        next_volatile() ;
                        result->status = NFS3_OK ;
                    }
                DIRBUFFERUNLOCK ;
                break ;
            case 0:
                /* first pass */
                FS_dir( load_volatile, dp->pn ) ;
                result->status = NFS3_OK ;
                break ;
            default:
                result->status = NFS3ERR_BAD_COOKIE ;
                break ;
            }
            break ;
        case desc_file:
            result->status = NFS3ERR_NOTDIR ;
            break ;
        default:
            /* file-type (ram based) directory */
            result->status = NFS3_OK ;
            FS_dir( static_dir, dp->pn ) ;
            break ;
        }
    }
    result->READDIR3res_u.resok.reply.entries = (void *) dp->buffer ;
    result->READDIR3res_u.resok.reply.eof = eof ;
}

void nfsproc3_readdirplus_3_svc(READDIRPLUS3args * argp, struct datapasser_t * dp ) {
    READDIRPLUS3res * result = (READDIRPLUS3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->dir.data.data_val ;
    size_t sbuf = argp->maxcount ;
    int indx = 0 ;
    int eof = 1 ;
    int spacetogo;
    char * endadds ;
    int nentry ;

    /* Embedded function */
    /* Add an entry to the allocated buffer */
    /* structs at front, strings at end */
    /* Needs to be called under mutex protection */
    int AddDirEntry( char * name, cookie3 ck ) {
        char path[OW_FULLNAME_MAX+strlen(dp->pn->path)+1] ;
        entryplus3 * E = (entryplus3 *) dp->buffer ;
        int l = strlen(name) ;
        int ret = 0 ;
        struct parsedname pnentry ;
        
        pnentry.si = dp->pn->si ;
        strcpy( path, dp->pn->path ) ;
        strcat( path, name ) ;

        if ( FS_ParsedName( path, &pnentry ) ) return 0 ;
        
        DIRSCANLOCK;
            /* test C length */
            if ( dp->buffer + (indx+1)*sizeof(entryplus3) + l + 1 + sizeof(unfs3_fh_t) > endadds ) {
                ret = 1 ; /* buffer full */
            } else {
                /* test XDR length */
                spacetogo -= ENTRYPLUS_SIZE + ((l+3)/4)*4 + 64 ;
                if ( spacetogo < 0 ) {
                    ret = 1 ;
                } else {
                    if ( indx>0 ) E[indx-1].nextentry = &(E[indx]) ;
                    E[indx].cookie = ck ;
                    E[indx].fileid = UniqueNum() ;
                    E[indx].nextentry = 0 ;
                    get_post_attr( &(E[indx].name_attributes), &pnentry ) ;

                    /* make room for file name at end of buffer */
                    endadds -= l+1 ;
                    strcpy(endadds,name) ;
                    E[indx].name = endadds ;

                    /* make room for filehandle at end of buffer */
                    endadds -= sizeof(unfs3_fh_t) ; /* space for file handle */
                    if ( fh_create( endadds, &pnentry ) ) { /* problem -- cannot make a valid filehandle! */
                        E[indx].name_handle.handle_follows = FALSE ;
                    } else {
                        E[indx].name_handle.handle_follows = TRUE ;
                        E[indx].name_handle.post_op_fh3_u.handle.data.data_len = sizeof(unfs3_fh_t) ;
                        E[indx].name_handle.post_op_fh3_u.handle.data.data_val = endadds ;
                    }
                    
                    ++ indx ;
                }
            }
        DIRSCANUNLOCK;

        FS_ParsedName_destroy(&pnentry) ;
        return ret ;
    }

    /* Embedded function */
    /* Callback function to FS_dir */
    /* Prints this directory element (not the whole path) */
    void load_volatile( const struct parsedname * const pn2 ) {
        char extname[OW_FULLNAME_MAX+1] ;
        FS_DirName( extname, OW_FULLNAME_MAX+1, pn2 ) ;
        if ( AddDirEntry( extname, (cookie3) 1 ) ) {
            /* Buffer full, now store rest of entries */
            DIRBUFFERLOCK;
                if ( eof ) { /* first excess entry */
                cookie3 cook = ++vb.cookie ;
                memcpy( &(result->READDIRPLUS3res_u.resok.cookieverf), &cook, sizeof(cookie3) ) ;
                    if ( vb.argz ) {
                        free(vb.argz) ;
                        vb.argz = NULL ;
                        vb.argz_len = 0 ;
                        vb.lastused = 0 ;
                        eof = 0 ;
                    }
                }
                argz_add( &(vb.argz), &(vb.argz_len), extname ) ;
            DIRBUFFERLOCK;
        }
    }

    /* Embedded function */
    void next_volatile( void ) {
        char * next ;
        while ( (next=argz_next( vb.argz, vb.argz_len, vb.lastused)) ) {
            if ( AddDirEntry( next, (cookie3) 2 ) ) return ;
            vb.lastused = next ;
        }
        eof = 1 ;
    }

    /* Embedded function */
    /* Callback function to FS_dir */
    /* Prints this directory element (not the whole path) */
    void static_dir( const struct parsedname * const pn2 ) {
        char extname[OW_FULLNAME_MAX+1] ;
        if ( (++nentry > argp->cookie) && eof) {
            FS_DirName( extname, OW_FULLNAME_MAX+1, pn2 ) ;
            if ( AddDirEntry( extname, (cookie3) nentry ) ) eof = 0 ;
        }
    }
    
    bad_post_attr( &(result->READDIRPLUS3res_u.resfail.dir_attributes) ) ;
    if ( sbuf < 256 ) {
        result->status = NFS3ERR_TOOSMALL ;
        return ;
    }
    
    if ( sbuf>8192) sbuf=8192 ; /* limit buffer size */

    if ( (dp->buffer = malloc(sbuf)) == NULL ) {
        result->status = NFS3ERR_SERVERFAULT ;
        return ;
    }
    spacetogo = argp->maxcount - sizeof(READDIRPLUS3resok) ;
    endadds = &(dp->buffer[sbuf]) ;

    if ( fh_interpret( fh, dp->pn ) ) {
        result->status = NFS3ERR_STALE ;
    } else {
        get_post_attr( &(result->READDIRPLUS3res_u.resfail.dir_attributes), dp->pn );
        switch( dp->pn->desc ) {
        case desc_error:
            result->status = NFS3ERR_IO ;
            break ;
        case desc_alarmdir:
        case desc_rootdir:
        case desc_branchdir:
            /* changable directory -- load buffer, then store excess for next call */
            switch( argp->cookie ) {
            case 1:
                /* not first pass */
                DIRBUFFERLOCK;
                    if ( memcmp( argp->cookieverf, &(vb.cookie), sizeof(cookie3) ) ) {
                        result->status = NFS3ERR_BAD_COOKIE ;
                    } else {
                        eof = 0 ;
                        memcpy( &(result->READDIRPLUS3res_u.resok.cookieverf), &(vb.cookie), sizeof(cookie3) ) ;
                        next_volatile() ;
                        result->status = NFS3_OK ;
                    }
                DIRBUFFERUNLOCK ;
                break ;
            case 0:
                /* first pass */
                FS_dir( load_volatile, dp->pn ) ;
                result->status = NFS3_OK ;
                break ;
            default:
                result->status = NFS3ERR_BAD_COOKIE ;
                break ;
            }
            break ;
        case desc_file:
            result->status = NFS3ERR_NOTDIR ;
            break ;
        default:
            /* file-type (ram based) directory */
            result->status = NFS3_OK ;
            FS_dir( static_dir, dp->pn ) ;
            break ;
        }
        result->READDIRPLUS3res_u.resok.reply.entries = (void *) dp->buffer ;
        result->READDIRPLUS3res_u.resok.reply.eof = eof ;
    }
}

void nfsproc3_fsstat_3_svc(FSSTAT3args * argp, struct datapasser_t * dp ) {
    FSSTAT3res * result = (FSSTAT3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->fsroot.data.data_val ;
    
    if ( fh_interpret( fh, dp->pn ) ) {
        bad_post_attr( &(result->FSSTAT3res_u.resfail.obj_attributes) ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        memset( &(result->FSSTAT3res_u.resok), 0, sizeof(FSSTAT3resok) ) ;
        get_post_attr( &(result->FSSTAT3res_u.resfail.obj_attributes), dp->pn ) ;
        result->status = NFS3_OK ;
    }
}

void nfsproc3_fsinfo_3_svc(FSINFO3args * argp, struct datapasser_t * dp ) {
    FSINFO3res * result = (FSINFO3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->fsroot.data.data_val ;

    if ( fh_interpret( fh, dp->pn ) ) {
        bad_post_attr( &(result->FSINFO3res_u.resfail.obj_attributes) ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        get_post_attr( &(result->FSINFO3res_u.resok.obj_attributes), dp->pn ) ;
        result->status = NFS3_OK;
        result->FSINFO3res_u.resok.rtmax = 16384;
        result->FSINFO3res_u.resok.rtpref = 4096;
        result->FSINFO3res_u.resok.rtmult = 4096;
        result->FSINFO3res_u.resok.wtmax = 16384;
        result->FSINFO3res_u.resok.wtpref = 4096;
        result->FSINFO3res_u.resok.wtmult = 4096;
        result->FSINFO3res_u.resok.dtpref = 4096;
        result->FSINFO3res_u.resok.maxfilesize = 64000;
        result->FSINFO3res_u.resok.time_delta.seconds = 1;
        result->FSINFO3res_u.resok.time_delta.nseconds = 0;
        result->FSINFO3res_u.resok.properties = FSF3_HOMOGENEOUS ;
    }
}

void nfsproc3_pathconf_3_svc(PATHCONF3args * argp, struct datapasser_t * dp ) {
    PATHCONF3res * result = (PATHCONF3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->object.data.data_val ;

    if ( fh_interpret( fh, dp->pn ) ) {
        bad_post_attr( &(result->PATHCONF3res_u.resfail.obj_attributes) ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        get_post_attr( &(result->PATHCONF3res_u.resok.obj_attributes), dp->pn ) ;
        result->PATHCONF3res_u.resok.linkmax = 0xFFFFFFFF;
        result->PATHCONF3res_u.resok.name_max = OW_FULLNAME_MAX;
        result->PATHCONF3res_u.resok.no_trunc = TRUE;
        result->PATHCONF3res_u.resok.chown_restricted = FALSE;
        result->PATHCONF3res_u.resok.case_insensitive = FALSE;
        result->PATHCONF3res_u.resok.case_preserving = TRUE;
        result->status = NFS3_OK ;
    }
}

void nfsproc3_commit_3_svc(COMMIT3args * argp, struct datapasser_t * dp ) {
    COMMIT3res * result = (COMMIT3res *) dp->result ;
    unfs3_fh_t *fh = (void *) argp->file.data.data_val ;
    pre_op_attr attr ;
    
    if ( fh_interpret( fh, dp->pn ) ) {
        bad_pre_attr( &attr ) ;
        result->status = NFS3ERR_STALE ;
    } else {
        get_pre_attr( &attr, dp->pn );
        result->status = NFS3_OK ;
    }
    memcpy( &(result->COMMIT3res_u.resok.file_wcc), &(attr.pre_op_attr_u.attributes), sizeof(wcc_attr) ) ;
}
